package main

import (
	"bytes"
	"errors"
	"fmt"
	"html/template"
	"io"
	"net/http"
	"strconv"
	"strings"
	"sync"

	log "github.com/sirupsen/logrus"

	"git.c3pb.de/gbe/invinoveritas/auth"
)

func (h Handler) img() http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		if r.Method != "GET" {
			httpError(w, "bad method", errors.New(r.Method), http.StatusMethodNotAllowed)
			return
		}

		args := r.URL.Query()

		if args.Get("id") == "" {
			httpError(w, "bad request", errors.New("empty id"), http.StatusBadRequest)
			return
		}

		id, err := strconv.Atoi(args.Get("id"))
		if err != nil {
			httpError(w, "can't parse id", err, http.StatusBadRequest)
			return
		}

		img, err := LoadPictureData(r.Context(), h.sqlDB, id)
		if err != nil {
			httpError(w, "can't load image", err, http.StatusNotFound)
			return
		}

		w.Header().Set("Content-Type", "image/png")

		_, err = io.Copy(w, bytes.NewBuffer(img))
		if err != nil {
			log.WithError(err).
				Error("can't write image")
			return
		}
	}
}

func (h Handler) detailsPost(w http.ResponseWriter, r *http.Request) {
	action := r.FormValue("action")
	switch action {
	case "save", "delete":
	default:
		httpError(w, "invalid parameters", fmt.Errorf("action missing or unknown: %q", action), http.StatusBadRequest)
		return
	}

	id, err := strconv.Atoi(r.FormValue("id"))
	if err != nil {
		httpError(w, "can't parse id", err, http.StatusBadRequest)
		return
	}

	if action == "delete" {
		err = DeleteVino(r.Context(), h.sqlDB, id)
		if err != nil {
			httpError(w, "can't delete wine", err, http.StatusInternalServerError)
			return
		}

		http.Redirect(w, r, "/", http.StatusSeeOther)
		return
	}

	v, err := LoadVino(r.Context(), h.sqlDB, id)
	if err != nil {
		// Only log error here, just create a new entry.
		log.WithError(err).
			WithField("id", id).
			Error("can't load vino")
	}

	name := r.FormValue("name")
	if name == "" {
		httpError(w, "name can't be empty", nil, http.StatusBadRequest)
		return
	}

	rating, err := strconv.Atoi(r.FormValue("rating"))
	if err != nil {
		httpError(w, "can't parse rating", nil, http.StatusBadRequest)
		return
	}

	country, err := ISO2CountryCodeFromString(r.FormValue("country"))
	if err != nil {
		httpError(w, "can't parse country", nil, http.StatusBadRequest)
		return
	}

	v.Name = name
	v.Rating = rating
	v.Country = country

	comment := strings.TrimSpace(r.FormValue("comment"))
	if comment != "" {
		if len(comment) > 1024 {
			comment = comment[:1024] + "..."
		}

		err = v.StoreComment(r.Context(), h.sqlDB, comment)
		if err != nil {
			httpError(w, "can't add comment", err, http.StatusInternalServerError)
			return
		}
	}

	err = r.ParseMultipartForm(16 * 1024 * 1024)
	if err != nil {
		httpError(w, "can't parse multipart form data", err, http.StatusInternalServerError)
		return
	}

	// See if there's an image file uploaded
	if len(r.MultipartForm.File["picture"]) != 0 {
		picFile, picHdr, err := r.FormFile("picture")
		if err != nil {
			httpError(w, "can't open picture file", err, http.StatusInternalServerError)
			return
		}
		defer picFile.Close()

		err = v.AddPicture(picFile, picHdr.Header.Get("Content-Type"))
		if err != nil {
			httpError(w, "can't load picture", err, http.StatusBadRequest)
			return
		}
	}

	err = v.Store(r.Context(), h.sqlDB)
	if err != nil {
		httpError(w, "can't store wine", err, http.StatusInternalServerError)
		return
	}

	http.Redirect(w, r, "/details?id="+r.FormValue("id"), http.StatusSeeOther)
	return
}

func (h Handler) details() http.Handler {
	var (
		once sync.Once
		tpl  *template.Template
	)

	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		once.Do(func() {
			tpl = template.Must(template.ParseFS(templateFS, "templates/base.tpl", "templates/details.tpl"))
		})

		if r.Method == "POST" {
			h.detailsPost(w, r)
			return
		}

		if r.Method != "GET" {
			httpError(w, "bad method", errors.New(r.Method), http.StatusMethodNotAllowed)
			return
		}

		args := r.URL.Query()

		if args.Get("id") == "" {
			httpError(w, "bad request", errors.New("empty id"), http.StatusBadRequest)
			return
		}

		id, err := strconv.Atoi(args.Get("id"))
		if err != nil {
			httpError(w, "can't parse id", err, http.StatusBadRequest)
			return
		}

		v, err := LoadVino(r.Context(), h.sqlDB, id)
		if err != nil {
			httpError(w, "can't load wine data", err, http.StatusInternalServerError)
			return
		}

		data := struct {
			Wine Vino
			User auth.User
		}{
			Wine: v,
			User: auth.Get(r),
		}

		err = tpl.ExecuteTemplate(w, "details.tpl", data)
		if err != nil {
			log.WithError(err).
				Error("can't execute details template")
			return
		}
	})
}
