package main

import (
	"errors"
	"fmt"
	"html/template"
	"net/http"

	log "github.com/sirupsen/logrus"

	"git.c3pb.de/gbe/invinoveritas/auth"
)

var userTemplate = template.Must(template.ParseFS(templateFS, "templates/base.tpl", "templates/user.tpl"))

func (h Handler) userCreate(w http.ResponseWriter, r *http.Request) (string, error) {
	name := r.PostForm.Get("new-user")
	if name == "" {
		return "", errors.New("name can't be empty")
	}

	if len(name) > 80 {
		return "", fmt.Errorf("user name too long %q", name)
	}

	generatedPassword, err := h.ap.createUser(r.Context(), name)
	if err != nil {
		return "", err
	}

	return generatedPassword, nil
}

func (h Handler) user() http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		user := auth.Get(r)

		data := struct {
			User       auth.User
			AllUsers   []string // User names
			FormErrors map[string]error
			Success    bool
			ShowResult bool

			GeneratedPassword string // For new users
			CreatedUser       string // Name for new users
		}{
			User:       user,
			FormErrors: make(map[string]error),
		}

		if r.Method == "GET" {
			var err error

			data.AllUsers, err = h.ap.listUsers(r.Context())
			if err != nil {
				httpError(w, "can't get list of all users", nil, http.StatusInternalServerError)
				return
			}

			err = userTemplate.ExecuteTemplate(w, "user.tpl", data)
			if err != nil {
				log.WithError(err).
					Error("can't execute details template")
			}

			return
		}

		if r.Method != "POST" {
			httpError(w, "bad method", errors.New(r.Method), http.StatusMethodNotAllowed)
			return
		}

		err := r.ParseForm()
		if err != nil {
			httpError(w, "can't parse POST form", err, http.StatusInternalServerError)
			return
		}

		defer func() {
			var err error

			data.AllUsers, err = h.ap.listUsers(r.Context())
			if err != nil {
				httpError(w, "can't get list of all users", nil, http.StatusInternalServerError)
				return
			}

			err = userTemplate.ExecuteTemplate(w, "user.tpl", data)
			if err != nil {
				log.WithError(err).
					Error("can't execute details template")
			}
		}()

		switch r.PostForm.Get("action") {
		case "update":
			data.ShowResult = true
		case "create":
			generatedPw, err := h.userCreate(w, r)
			if err != nil {
				data.FormErrors["NewUser"] = err
			}

			data.GeneratedPassword = generatedPw
			data.CreatedUser = r.PostForm.Get("new-user")

			log.WithField("name", data.CreatedUser).
				Info("user created")

			return
		case "delete":
			err := h.ap.deleteUser(r.Context(), r.PostForm.Get("username"))
			if err != nil {
				data.FormErrors["DeleteUser"] = err
			}

			return
		default:
			httpError(w, "unknown action", nil, http.StatusBadRequest)
			return
		}

		if r.PostForm.Get("new-pw") != r.PostForm.Get("new-pw-repeat") {
			data.FormErrors["NewPW"] = errors.New("new passwords don't match")
			return
		}

		if len(r.PostForm.Get("new-pw")) == 0 {
			data.FormErrors["NewPW"] = errors.New("new password can't be empty")
			return
		}

		err = h.ap.updatePassword(r.Context(), user.Name, r.PostForm.Get("old-pw"), r.PostForm.Get("new-pw"))
		if err != nil {
			data.FormErrors["OldPW"] = err
			return
		}

		data.Success = true
	})
}
